# MetroPSF (C) Copyright 2021, Maxym Usatov <maxim.usatov@bcsatellite.net>
# Please refer to metropsf.pdf for license information.

import numpy as np
import os.path
import numbers
import sys
import pandas as pd
from astropy.io import fits
from astropy import units as u
from astropy.wcs import WCS
from astropy.coordinates import SkyCoord
from PIL import Image, ImageTk, ImageMath
from astroquery.astrometry_net import AstrometryNet
from astroquery.vizier import Vizier
import matplotlib
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
from matplotlib import cm
from mpl_toolkits.mplot3d import Axes3D
from astropy.nddata import Cutout2D

import tkinter as tk
from tkinter import filedialog as fd

# Photometry
from photutils.detection import IRAFStarFinder
from photutils.psf import IntegratedGaussianPRF, DAOGroup
from photutils.background import MMMBackground, MADStdBackgroundRMS
from photutils.psf import IterativelySubtractedPSFPhotometry
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.stats import gaussian_sigma_to_fwhm
from astropy.stats import SigmaClip
from photutils import Background2D, MedianBackground

# No imports beyond this line.

def save_background_image(stretch_min, stretch_max, zoom_level, image_data):
    global FITS_minimum
    global FITS_maximum
    global background_image
    background_image = Image.fromarray(image_data)
    width, height = background_image.size
    new_size = (int(width * zoom_level), int(height * zoom_level))
    background_image = background_image.resize(new_size, Image.ANTIALIAS)
    background_image = ImageMath.eval("(a + " + str(stretch_min / 100 * FITS_maximum) + ") * 255 / " + str(stretch_max / 100 * FITS_maximum), a=background_image)
    background_image = ImageMath.eval("convert(a, 'L')", a=background_image)
    background_image.save('background.jpg')


def generate_FITS_thumbnail(stretch_min, stretch_max, zoom_level):
    global generated_image
    global image_data
    global FITS_minimum
    global FITS_maximum
    generated_image = Image.fromarray(image_data)
    width, height = generated_image.size
    new_size = (int(width * zoom_level), int(height * zoom_level))
    generated_image = generated_image.resize(new_size, Image.ANTIALIAS)
    generated_image = ImageMath.eval("(a + " + str(stretch_min / 100 * FITS_maximum) + ") * 255 / " + str(stretch_max / 100 * FITS_maximum), a=generated_image)

image_file = ""

class MyGUI:

    astrometry_API_key = "pwjgdcpwaugkhkln"

    zoom_level = 1
    zoom_step = 0.5
    photometry_results_plotted = False
    results_tab_df=pd.DataFrame()
    bkg_value = 0
    fit_shape = 15
    histogram_slider_low = 0
    histogram_slider_high = 5
    last_clicked_x = 0
    last_clicked_y = 0
    a = 0
    b = 0
    image_file = ""

    def console_msg(self, message):
        self.console.insert(tk.END, message+"\n")
        self.console.see(tk.END)
        self.window.update_idletasks()

    def display_image(self):
        self.canvas.delete("all")
        global generated_image
        generate_FITS_thumbnail(self.histogram_slider_low, self.histogram_slider_high, self.zoom_level)
        self.image = ImageTk.PhotoImage(generated_image)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=self.image)
        self.canvas.config(scrollregion=self.canvas.bbox(tk.ALL))
        self.canvas.bind("<Button-1>", self.mouse_click)
        if self.photometry_results_plotted:
            self.plot_photometry()

    def display_background(self):
        self.canvas.delete("all")
        self.console_msg("Displaying extracted background.")
        global background_image
        self.image = ImageTk.PhotoImage(background_image)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=self.image)
        self.canvas.config(scrollregion=self.canvas.bbox(tk.ALL))


    def load_FITS(self, image_file):
        global image_figure
        global image_data
        global image
        global image_width
        global image_height
        global header
        global FITS_minimum
        global FITS_maximum
        global generated_image
        try:
            self.console_msg("Loading FITS: " + image_file)
            with fits.open(image_file) as image:
                self.image_file = image_file
                self.filename_label['text'] = "FITS: " + image_file
                self.canvas.delete("all")
                self.zoom_level = 1
                self.photometry_results_plotted = False
                header = image[0].header
                image_data = fits.getdata(image_file)
                image_width = image_data.shape[1]
                image_height = image_data.shape[0]
                self.wcs_header = WCS(image[0].header)
                self.console_msg("WCS Header: " + str(self.wcs_header))
                FITS_minimum = np.min(image_data)
                FITS_maximum = np.max(image_data)
                self.console_msg("Width: " + str(image_width) + " Height: " + str(image_height))
                self.console_msg("FITS Minimum: " + str(FITS_minimum) + " Maximum: " + str(FITS_maximum))
                if 'filter' in header:
                    self.filter = str(header['filter'])
                    self.set_entry_text(self.filter_entry, self.filter)
                    self.console_msg("Filter: " + self.filter)
                else:
                    self.console_msg("Filter name not in FITS header. Set filter manually.")
                if 'exptime' in header:
                    exptime = header['exptime']
                    self.set_entry_text(self.exposure_entry, exptime)
                    self.console_msg("Exposure: " + str(exptime))
                else:
                    self.console_msg("Exposure (EXPTIME) not in FITS header. Set exposure manually.")

                self.bkg_value = np.median(image_data)
                self.console_msg("Median background level, ADU: " + str(self.bkg_value))
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            self.console_msg(str(exc_tb.tb_lineno)+" "+str(e))
            pass

    def initialize_debug(self):
        image_file = "123.fits"
        if len(image_file) > 0:
            self.load_FITS(image_file)
            self.display_image()


    def open_FITS_file(self):
        image_file = fd.askopenfilename()
        if len(image_file) > 0:
            self.load_FITS(image_file)
            self.display_image()

    def save_FITS_file_as(self):
        global image
        global image_data
        global header
        file_name = fd.asksaveasfile(defaultextension='.fits')
        try:
            if len(str(file_name)) > 0:
                self.console_msg("Saving FITS as " + str(file_name.name))
                fits.writeto(file_name.name, image_data, header, overwrite=True)
                self.console_msg("Saved.")
        except Exception as e:
            self.console_msg(str(e))

    def save_FITS_file(self):
        global image
        global image_data
        global header
        file_name = self.image_file
        try:
            if len(str(file_name)) > 0:
                self.console_msg("Saving FITS as " + str(file_name))
                fits.writeto(file_name, image_data, header, overwrite=True)
                self.console_msg("Saved.")
        except Exception as e:
            self.console_msg(str(e))

    def perform_photometry(self):
        self.console_msg("Performing interactively subtracted PSF photometry..")
        try:
            self.fit_shape = int(self.photometry_aperture_entry.get())
            fwhm = int(self.fwhm_entry.get())
            star_detection_threshold = float(self.star_detection_threshold_entry.get())
            iterations = int(self.photometry_iterations_entry.get())
            bkgrms = MADStdBackgroundRMS()
            sharplo = float(self.sharplo_entry.get())
            bkg_filter_size = int(self.bkg_filter_size_entry.get())
            std = bkgrms(image_data)
            self.console_msg("Finding stars..")
            iraffind = IRAFStarFinder(threshold=star_detection_threshold * std,
                                      fwhm=fwhm, roundhi=3.0, roundlo=-5.0,
                                       sharplo=sharplo, sharphi=2.0)

            daogroup = DAOGroup(2.0 * fwhm * gaussian_sigma_to_fwhm)
            #mmm_bkg = MMMBackground()

            sigma_clip = SigmaClip(sigma = 3.0)
            bkg_estimator = MedianBackground()
            self.console_msg("Estimating background..")
            bkg = Background2D(image_data, (self.fit_shape * 1, self.fit_shape * 1),
                               filter_size=(bkg_filter_size, bkg_filter_size), sigma_clip=sigma_clip,
                               bkg_estimator=bkg_estimator)
            clean_image = image_data-bkg.background
            save_background_image(self.histogram_slider_low, self.histogram_slider_high, self.zoom_level, bkg.background)

            fitter = LevMarLSQFitter()
            psf_model = IntegratedGaussianPRF(sigma=2)  # sigma=2 here is the initial guess
            psf_model.sigma.fixed = False   # This allows to fit Gaussian PRF sigma as well
            photometry = IterativelySubtractedPSFPhotometry(finder=iraffind,
                                                            group_maker=daogroup,
                                                            psf_model=psf_model,
                                                            bkg_estimator=None,
                                                            fitter=LevMarLSQFitter(),
                                                            niters=iterations, fitshape=(self.fit_shape, self.fit_shape))
            self.console_msg("Performing photometry..")
            result_tab = photometry(image=clean_image)
            residual_image = photometry.get_residual_image()

            self.console_msg("Done. PSF fitter message(s): " + str(fitter.fit_info['message']))

            self.results_tab_df = result_tab.to_pandas()
            self.results_tab_df.to_csv(self.image_file + ".phot", index=False)
            self.console_msg("Photometry saved to " + str(self.image_file + ".phot"))
            self.plot_photometry()
        except Exception as e:
            self.console_msg(str(e))


    def create_circle(self, x, y, r, canvas_name, outline_color="grey50"):  # center coordinates, radius
        x0 = x - r
        y0 = y - r
        x1 = x + r
        y1 = y + r
        return canvas_name.create_oval(x0, y0, x1, y1, outline=outline_color)

    def plot_photometry(self):
        try:
            matches_in_photometry_table = False
            if "match_id" in self.results_tab_df:
                matches_in_photometry_table = True
            exptime = float(self.exposure_entry.get())
            if os.path.isfile(self.image_file+".phot"):
                self.fit_shape = int(self.photometry_aperture_entry.get())
                self.results_tab_df = pd.read_csv(self.image_file + ".phot")
                # Calculate instrumental magnitudes
                self.results_tab_df["inst_mag"] = -2.5 * np.log(self.results_tab_df["flux_fit"] / exptime) + 25
                self.photometry_results_plotted = True
                for index, row in self.results_tab_df.iterrows():
                    outline_color = "grey50"
                    if matches_in_photometry_table:
                        if len(str(row["match_id"])) > 0 and str(row["match_id"]) != "nan":
                            outline_color="green"
                    self.create_circle(x=row["x_fit"] * self.zoom_level, y=row["y_fit"] * self.zoom_level,
                                       r=self.fit_shape / 2 * self.zoom_level, canvas_name=self.canvas, outline_color=outline_color)
        except Exception as e:
            self.console_msg(str(e))

    def hide_photometry(self):
        self.photometry_results_plotted = False
        self.display_image()

    def match_photometry_table(self, x, y, r=5):
        x_criterion = self.results_tab_df['x_fit'] < (x + r)
        matched_objects = self.results_tab_df[x_criterion]
        x_criterion = self.results_tab_df['x_fit'] > (x - r)
        matched_objects = matched_objects[x_criterion]
        y_criterion = self.results_tab_df['y_fit'] < (y + r)
        matched_objects = matched_objects[y_criterion]
        y_criterion = self.results_tab_df['y_fit'] > (y - r)
        matched_objects = matched_objects[y_criterion]
        if len(matched_objects) > 0:
            return(matched_objects.iloc[0]["x_fit"],
                   matched_objects.iloc[0]["y_fit"],
                   matched_objects.iloc[0]["flux_fit"],
                   matched_objects.iloc[0]["sigma_fit"],
                   matched_objects.iloc[0]["inst_mag"])
        else:
            return(0, 0, 0, 0, 0)

    def mouse_click(self, event):
        global image_data
        self.console_msg("")
        x = int(self.canvas.canvasx(event.x) / self.zoom_level)
        y = int(self.canvas.canvasy(event.y) / self.zoom_level)
        self.last_clicked_x = x
        self.last_clicked_y = y
        ADU = image_data[y-1, x-1]
        sky = self.wcs_header.pixel_to_world(x, y)
        sky_coordinate_string = ""
        if hasattr(sky, 'ra'):
            c = SkyCoord(ra=sky.ra, dec=sky.dec)
            sky_coordinate_string = "α δ: "+c.to_string("hmsdms")
        self.console_msg("Click X: "+str(x)+"\t Y: "+str(y)+"\t ADU: "+str(ADU) + "\t\t\t" + sky_coordinate_string)
        psf_canvas_x = x
        psf_canvas_y = y
        if self.photometry_results_plotted:
            x_fit, y_fit, flux_fit, sigma_fit, inst_mag = self.match_photometry_table(x, y)
            sky = self.wcs_header.pixel_to_world(x_fit, y_fit)
            sky_coordinate_string = ""
            if hasattr(sky, 'ra'):
                c = SkyCoord(ra=sky.ra, dec=sky.dec)
                sky_coordinate_string = " α δ: " + c.to_string("hmsdms")
                if x_fit != 0 and y_fit != 0:
                    psf_canvas_x = x_fit
                    psf_canvas_y = y_fit
            self.console_msg("Photometry fits, X: " + str(round(x_fit,2)) + " Y: " + str(round(y_fit,2)) + " Flux (ADU): " + str(
                round(flux_fit,2)) + " Instrumental magnitude: " + str(round(inst_mag,2)) + " PSF σ: " + str(round(sigma_fit, 2)) +sky_coordinate_string)
            if "match_id" in self.results_tab_df:
                matching_star_criterion = (self.results_tab_df["x_fit"] == x_fit) & (self.results_tab_df["y_fit"] == y_fit)
                if len(self.results_tab_df[matching_star_criterion]) > 0:
                    matching_star = self.results_tab_df[matching_star_criterion].iloc[0]
                    if type(matching_star["match_id"]) in (str, int, np.float64):
                        self.console_msg(
                            "Matching catalog source ID: " + str(matching_star["match_id"]) + " magnitude: " + str(
                                matching_star["match_mag"]))
            if self.a != 0 and self.b != 0 and x_fit != 0 and y_fit != 0:
                differential_magnitude = round(inst_mag * self.a + self.b, 3)
                snr = flux_fit/self.bkg_value
                poisson_noise_error = 1.0857 / snr
                total_error = np.nan
                if "mag_error" in self.results_tab_df:
                    linreg_error = self.results_tab_df["mag_error"].mean()
                    total_error = round(np.sqrt(linreg_error ** 2 + poisson_noise_error ** 2), 2)
                    self.console_msg("Poisson Noise Error: " + str(round(poisson_noise_error, 3)) + " SNR: " + str(
                        round(snr, 3)) + " Linear Regression Error: " + str(round(linreg_error, 3)))
                self.console_msg(
                        "Differential ensemble photometry magnitude: " + str(differential_magnitude) + " ± " + str(
                            total_error))
                # Checking PSF sigma
                median_psf_sigma = self.results_tab_df["sigma_fit"].median()
                if sigma_fit > median_psf_sigma*1.2:
                    self.console_msg("WARNING: This source has PSF σ exceeding median value by more than 20%. Check PSF shape - likely non-linear sensor regime or improperly subtracted sky background.")
        self.update_PSF_canvas(psf_canvas_x, psf_canvas_y)

    def update_PSF_canvas(self, x, y):
        global image_data
        global FITS_minimum
        global FITS_maximum
        self.fit_shape = int(self.photometry_aperture_entry.get())
        x0 = int(x - (self.fit_shape - 1) / 2)
        y0 = int(y - (self.fit_shape - 1) / 2)
        x1 = int(x + (self.fit_shape - 1) / 2)
        y1 = int(y + (self.fit_shape - 1) / 2)
        position=(x, y)
        size=(self.fit_shape-1, self.fit_shape-1)
        data = Cutout2D(image_data, position, size).data
        x = np.arange(x0, x1, 1)
        y = np.arange(y0, y1, 1)
        x, y = np.meshgrid(x, y)
        self.psf_plot.clear()
        self.psf_plot.plot_surface(x, y, data, cmap=cm.jet)
        self.psf_plot_canvas.draw()

    def update_PSF_canvas_2d(self, x, y):
        global image_data
        global FITS_minimum
        global FITS_maximum
        image_crop = Image.fromarray(image_data)
        self.fit_shape = int(self.photometry_aperture_entry.get())
        x0 = int(x - (self.fit_shape-1)/2)
        y0 = int(y - (self.fit_shape-1)/2)
        x1 = int(x + (self.fit_shape-1)/2)
        y1 = int(y + (self.fit_shape-1)/2)
        image_crop = image_crop.crop((x0, y0, x1, y1))
        image_crop = image_crop.resize((300, 300), resample=0)
        image_crop = ImageMath.eval("a * 255 / " + str(self.histogram_slider_high / 100 * FITS_maximum), a=image_crop)
        self.image_crop = ImageTk.PhotoImage(image_crop)
        self.psf_canvas.create_image(0, 0, anchor=tk.NW, image=self.image_crop)


    def zoom_in(self):
        self.canvas.scale("all", 0, 0, 1+self.zoom_step, 1+self.zoom_step)
        self.zoom_level = self.zoom_level * (1+self.zoom_step)
        self.console_msg("Zoom: "+str(self.zoom_level))
        self.display_image()

    def zoom_out(self):
        self.canvas.scale("all", 0, 0, 1-self.zoom_step, 1-self.zoom_step)
        self.zoom_level = self.zoom_level * (1-self.zoom_step)
        self.console_msg("Zoom: " + str(self.zoom_level))
        self.display_image()

    def zoom_100(self):
        self.canvas.scale("all", 0, 0, 1, 1)
        self.zoom_level = 1
        self.console_msg("Zoom: " + str(self.zoom_level))
        self.display_image()

    def solve_image(self):
        global generated_image
        global header
        self.console_msg("Solving via Astrometry.Net..")
        try:
            ast = AstrometryNet()
            ast.api_key = self.astrometry_API_key
            sources_df = self.results_tab_df.sort_values("flux_fit", ascending=False)
            width, height = generated_image.size

            self.wcs_header = ast.solve_from_source_list(sources_df['x_fit'], sources_df['y_fit'],
                                                         width, height,
                                                         solve_timeout=120)
            self.console_msg(
                "Astrometry.Net solution reference point RA: " + str(self.wcs_header["CRVAL1"]) + " Dec: " + str(
                    self.wcs_header["CRVAL2"]))
            header = header + self.wcs_header
            self.wcs_header = WCS(header)
        except Exception as e:
            self.console_msg(str(e))

    def get_comparison_stars(self):
        global image_width, image_height

        self.filter = self.filter_entry.get()

        frame_center = self.wcs_header.pixel_to_world(int(image_width/2), (image_height/2))
        frame_center_coordinates = SkyCoord(ra=frame_center.ra, dec=frame_center.dec)
        frame_edge = self.wcs_header.pixel_to_world(int(image_width), (image_height/2))
        frame_edge_coordinates = SkyCoord(ra=frame_edge.ra, dec=frame_edge.dec)
        frame_radius = frame_edge.separation(frame_center)
        self.console_msg("Inquiring VizieR, center α δ "+frame_center.to_string("hmsdms")+", radius "+str(frame_radius))

        if self.catalog_stringvar.get() == "APASS DR9":
            catalog = "II/336"

        if self.catalog_stringvar.get() == "URAT1":
            catalog = "I/329"

        if self.catalog_stringvar.get() == "USNO-B1.0":
            catalog = "I/284"

        if self.catalog_stringvar.get() == "VizieR Catalog":
            catalog = self.vizier_catalog_entry.get()

        vizier_mag_column = self.filter + "mag"

        comparison_stars = Vizier(catalog=catalog, row_limit=-1).query_region(frame_center, frame_radius)[0]
        self.console_msg("Found "+str(len(comparison_stars))+" objects in the field.")
        #print(comparison_stars)
        if vizier_mag_column not in comparison_stars.colnames:
            self.console_msg("Catalog "+self.catalog_stringvar.get()+" does not list "+self.filter+" magnitudes.")
            return

        if True:
            self.console_msg("Updating photometry table with sky coordinates..")
            for index, row in self.results_tab_df.iterrows():
                sky = self.wcs_header.pixel_to_world(row["x_fit"], row["y_fit"])
                c = SkyCoord(ra=sky.ra, dec=sky.dec)
                self.results_tab_df.loc[index, "ra_fit"] = c.ra / u.deg
                self.results_tab_df.loc[index, "dec_fit"] = c.dec / u.deg
            self.results_tab_df.to_csv(self.image_file+".phot", index=False)

        if True:
            self.console_msg("Matching catalogs..")
            catalog_comparison = SkyCoord(comparison_stars["RAJ2000"], comparison_stars["DEJ2000"])
            for index, row in self.results_tab_df.iterrows():
                photometry_star_coordinates = SkyCoord(ra=row["ra_fit"] * u.deg, dec=row["dec_fit"] * u.deg)
                match_index, d2d_match, d3d_match = photometry_star_coordinates.match_to_catalog_sky(catalog_comparison)
                print(str(photometry_star_coordinates))
                print(match_index)
                match_id = comparison_stars[match_index][0]  # Name of the catalog
                match_ra = comparison_stars[match_index]["RAJ2000"]
                match_dec = comparison_stars[match_index]["DEJ2000"]
                match_mag = comparison_stars[match_index][vizier_mag_column]
                match_coordinates = SkyCoord(ra=match_ra * u.deg, dec=match_dec * u.deg)
                separation = photometry_star_coordinates.separation(match_coordinates)
                if separation < 0.000555556 * u.deg:
                    self.results_tab_df.loc[index, "match_id"] = str(self.catalog_stringvar.get()) + " " + str(match_id)
                    self.results_tab_df.loc[index, "match_ra"] = match_ra
                    self.results_tab_df.loc[index, "match_dec"] = match_dec
                    self.results_tab_df.loc[index, "match_mag"] = match_mag
                else:
                    self.results_tab_df.loc[index, "match_id"] = ""
                    self.results_tab_df.loc[index, "match_ra"] = ""
                    self.results_tab_df.loc[index, "match_dec"] = ""
                    self.results_tab_df.loc[index, "match_mag"] = ""
            self.results_tab_df.to_csv(self.image_file + ".phot", index=False)
            self.console_msg("Photometry table saved to " + str(self.image_file + ".phot"))
            self.update_display()

    def set_entry_text(self, entry, text):
        entry.delete(0, tk.END)
        entry.insert(0, text)

    def update_display(self):
        self.display_image()
        self.update_PSF_canvas(self.last_clicked_x, self.last_clicked_y)

    def safe_float_convert(self, x):
        try:
            z = float(x)
            if np.isnan(z):
                return False    # Nan!
            return True  # numeric, success!
        except ValueError:
            return False  # not numeric
        except TypeError:
            return False  # null type

    def find_linear_regression_model(self):
        # Select stars with catalog matches with known magnitudes
        self.console_msg("Finding linear regression model with ensemble stars..")
        try:
            min_comparison_magnitude = int(self.min_ensemble_magnitude_entry.get())
            max_comparison_magnitude = int(self.max_ensemble_magnitude_entry.get())

            # Only stars with known magnitude from catalogs
            mask = self.results_tab_df['match_mag'].map(self.safe_float_convert)
            ensemble_stars = self.results_tab_df.loc[mask]

            # Convert magnitudes to floats
            ensemble_stars["inst_mag"] = ensemble_stars["inst_mag"].astype(float)
            ensemble_stars["match_mag"] = ensemble_stars["match_mag"].astype(float)

            # Filter by magnitude
            mask = ensemble_stars['match_mag'] < max_comparison_magnitude
            ensemble_stars = ensemble_stars.loc[mask]
            mask = ensemble_stars['match_mag'] > min_comparison_magnitude
            ensemble_stars = ensemble_stars.loc[mask]

            x = ensemble_stars["inst_mag"]
            y = ensemble_stars["match_mag"]

            weights = None

            if self.weighting_stringvar.get() == "Raw Flux":
                weights = ensemble_stars["flux_0"]

            if self.weighting_stringvar.get() == "Instrumental Magnitude":
                weights = 1/ensemble_stars["inst_mag"]

            if self.weighting_stringvar.get() == "PSF Sigma":
                weights = 1/ensemble_stars["sigma_fit"]

            self.a, self.b = np.polyfit(x, y, deg=1, w=weights)  # Fit a 1st degree polynomial
            fit_fn = np.poly1d((self.a, self.b))
            yhat = fit_fn(x)
            ybar = np.sum(y) / len(y)
            ssreg = np.sum((yhat - ybar) ** 2)
            sstot = np.sum((y - ybar) ** 2)
            r_squared = ssreg / sstot
            self.console_msg(
                "Linear regression fit: a = " + str(self.a) + " b = " + str(self.b) + " r^2 = " + str(r_squared))

            self.linreg_plot.clear()
            self.linreg_plot.plot(x, y, 'ro', ms=1)
            self.linreg_plot.plot(x, fit_fn(x), '--k', ms=1)
            self.plot_canvas.draw()

            for index, row in self.results_tab_df.iterrows():
                self.results_tab_df.loc[index, "differential_mag"] = round(float(row["inst_mag"]) * self.a + self.b, 3)
            for index, row in self.results_tab_df.iterrows():
                try:
                    self.results_tab_df.loc[index, "mag_error"] = abs(
                        float(row["differential_mag"]) - float(row["match_mag"]))
                except:
                    continue
            self.results_tab_df.to_csv(self.image_file + ".phot", index=False)
            self.console_msg("Photometry table with differential magnitudes saved to " + str(self.image_file + ".phot"))
        except Exception as e:
            self.console_msg(str(e))

    def delete_photometry_file(self):
        if os.path.isfile(self.image_file + ".phot"):
            os.remove(self.image_file + ".phot")
            self.console_msg("Photometry file deleted.")
            self.a = 0
            self.b = 0
            self.update_display()

    def update_histogram_low(self, value):
        self.histogram_slider_low = int(value)
        self.update_display()

    def update_histogram_high(self, value):
        self.histogram_slider_high = int(value)
        self.update_display()

    def __init__(self):
        self.window = tk.Tk()
        self.screen_width = self.window.winfo_screenwidth()
        self.screen_height = self.window.winfo_screenheight()

        # Matplotlib settings
        matplotlib.rc('xtick', labelsize=7)
        matplotlib.rc('ytick', labelsize=7)

        # Maximize that works everywhere
        m = self.window.maxsize()
        self.window.geometry('{}x{}+0+0'.format(*m))

        self.window.title("MetroPSF 0.1")

        self.menubar = tk.Menu(self.window)
        self.filemenu = tk.Menu(self.menubar, tearoff=0)
        self.filemenu.add_command(label="Open...", command=self.open_FITS_file)
        self.filemenu.add_command(label="Save", command=self.save_FITS_file)
        self.filemenu.add_command(label="Save As...", command=self.save_FITS_file_as)
        self.menubar.add_cascade(label="File", menu=self.filemenu)

        self.viewmenu = tk.Menu(self.menubar, tearoff=0)
        self.viewmenu.add_command(label="Update", command=self.update_display)
        self.viewmenu.add_command(label="Zoom In", command=self.zoom_in)
        self.viewmenu.add_command(label="Zoom Out", command=self.zoom_out)
        self.viewmenu.add_command(label="100% Zoom", command=self.zoom_100)
        self.menubar.add_cascade(label="View", menu=self.viewmenu)

        self.photometrymenu = tk.Menu(self.menubar, tearoff=0)
        self.photometrymenu.add_command(label="Iteratively Subtracted PSF Photometry", command=self.perform_photometry)
        self.photometrymenu.add_command(label="Plot", command=self.plot_photometry)
        self.photometrymenu.add_command(label="Hide", command=self.hide_photometry)
        self.photometrymenu.add_command(label="Solve Image", command=self.solve_image)
        self.photometrymenu.add_command(label="Get Comparison Stars", command=self.get_comparison_stars)
        self.photometrymenu.add_command(label="Find Regression Model", command=self.find_linear_regression_model)
        self.photometrymenu.add_command(label="Delete Photometry File", command=self.delete_photometry_file)
        self.photometrymenu.add_command(label="Display Background", command=self.display_background)
        self.menubar.add_cascade(label="Photometry", menu=self.photometrymenu)

        self.window.config(menu=self.menubar)

        self.left_half = tk.Frame(self.window)  # Left half of the window
        self.left_half.grid(row=0, column=0, sticky=tk.NSEW)
        self.center = tk.Frame(self.window)  # Center of the window
        self.center.grid(row=0, column=1, sticky=tk.NSEW)
        self.right_half = tk.Frame(self.window)  # Right half of the window
        self.right_half.grid(row=0, column=2, sticky=tk.NSEW)
        tk.Grid.columnconfigure(self.window, 1, weight=1) # Expand center horizontally
        tk.Grid.rowconfigure(self.window, 0, weight=1) # Expand everything vertically


        self.filename_label = tk.Label(self.center, text="FITS:" + image_file)
        self.filename_label.grid(row=0, column=0) # Place label

        self.canvas = tk.Canvas(self.center, bg='black') # Main canvas
        self.canvas.grid(row=1, column=0, sticky=tk.N+tk.S+tk.E+tk.W) # Place main canvas, sticky to occupy entire
                                                                      # cell dimensions
        tk.Grid.columnconfigure(self.center, 0, weight=1) # Expand main canvas column to fit whole window
        tk.Grid.rowconfigure(self.center, 1, weight=1) # Expand main canvas row to fit whole window
        self.canvas_scrollbar_V = tk.Scrollbar(self.center, orient=tk.VERTICAL) # Main canvas scrollbars
        self.canvas_scrollbar_V.grid(row=1, column=1)
        self.canvas_scrollbar_V.grid(sticky=tk.N+tk.S+tk.E+tk.W, column=1, row=1)
        self.canvas_scrollbar_H = tk.Scrollbar(self.center, orient=tk.HORIZONTAL)
        self.canvas_scrollbar_H.grid(row=2, column=0)
        self.canvas_scrollbar_H.grid(sticky=tk.N + tk.S + tk.E + tk.W, column=0, row=2)
        self.canvas_scrollbar_H.config(command=self.canvas.xview)
        self.canvas_scrollbar_V.config(command=self.canvas.yview)
        self.canvas.config(xscrollcommand=self.canvas_scrollbar_H.set)
        self.canvas.config(yscrollcommand=self.canvas_scrollbar_V.set)

        self.right_frame = tk.Frame(self.right_half) # We will lay out interface things into the new right_frame grid
        self.right_frame.grid(row=1, column=2, sticky=tk.N) # Place right_frame into the top of the main canvas row, right next to it
        #self.psf_canvas = tk.Canvas(self.right_frame, bg='grey', width=300, height=300) # Small PSF canvas
        self.fig_psf = Figure()
        self.psf_plot = self.fig_psf.add_subplot(111, projection='3d')
        self.psf_plot_canvas = FigureCanvasTkAgg(self.fig_psf, self.right_frame) # PSF 3D plot canvas - Matplotlib wrapper for Tk
        self.psf_plot_canvas.draw()
        self.psf_canvas = self.psf_plot_canvas.get_tk_widget()
        self.psf_canvas.config(width=int(self.screen_width/8.5), height=int(self.screen_width/8.5))
        self.psf_canvas.grid(row=0, column=0) # Allocate small PSF canvas to a new grid inside the right_frame

        self.fig = Figure()
        self.linreg_plot = self.fig.add_subplot(111)
        self.plot_canvas = FigureCanvasTkAgg(self.fig, self.right_frame) # Linear regression canvas - Matplotlib wrapper for Tk
        self.plot_canvas.draw()
        self.linreg_canvas = self.plot_canvas.get_tk_widget()
        self.linreg_canvas.config(width=int(self.screen_width/8.5), height=int(self.screen_width/12))
        self.linreg_canvas.grid(row=1, column=0) # Allocate small PSF canvas to a new grid inside the right_frame

        self.left_frame = tk.Frame(self.left_half) # We will lay out interface things into the new right_frame grid
        self.left_frame.grid(row=0, column=0, sticky=tk.N) # Place right_frame into the top of the main canvas row, right next to it

        self.settings_frame = tk.Frame(self.left_frame)  # Frame to hold settings grid
        self.settings_frame.grid(row=2, column=0, sticky=tk.NSEW)  # Settings_frame under the canvas in the right_frame
        tk.Grid.columnconfigure(self.settings_frame, 0, weight=1)  # Expand settings_frame column that holds labels

        settings_entry_width = 6

        self.photometry_aperture_label = tk.Label(self.settings_frame, text="Fitting Width/Height, px:")
        self.photometry_aperture_label.grid(row=0, column=0, sticky=tk.W)
        self.photometry_aperture_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.photometry_aperture_entry.grid(row=0, column=1, sticky=tk.E)
        self.set_entry_text(self.photometry_aperture_entry, self.fit_shape)

        self.min_ensemble_magnitude_label = tk.Label(self.settings_frame, text="Minimum Ensemble Magnitude:")
        self.min_ensemble_magnitude_label.grid(row=1, column=0, sticky=tk.W)
        self.min_ensemble_magnitude_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.min_ensemble_magnitude_entry.grid(row=1, column=1, sticky=tk.E)
        self.set_entry_text(self.min_ensemble_magnitude_entry, "7")

        self.max_ensemble_magnitude_label = tk.Label(self.settings_frame, text="Maximum Ensemble Magnitude:")
        self.max_ensemble_magnitude_label.grid(row=2, column=0, sticky=tk.W)
        self.max_ensemble_magnitude_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.max_ensemble_magnitude_entry.grid(row=2, column=1, sticky=tk.E)
        self.set_entry_text(self.max_ensemble_magnitude_entry, "20")

        self.fwhm_label = tk.Label(self.settings_frame, text="FWHM, px:")
        self.fwhm_label.grid(row=3, column=0, sticky=tk.W)
        self.fwhm_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.fwhm_entry.grid(row=3, column=1, sticky=tk.E)
        self.set_entry_text(self.fwhm_entry, "4")

        self.star_detection_threshold_label = tk.Label(self.settings_frame, text="Star Detection Threshold, σ:")
        self.star_detection_threshold_label.grid(row=4, column=0, sticky=tk.W)
        self.star_detection_threshold_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.star_detection_threshold_entry.grid(row=4, column=1, sticky=tk.E)
        self.set_entry_text(self.star_detection_threshold_entry, "10")

        self.photometry_iterations_label = tk.Label(self.settings_frame, text="Photometry Iterations:")
        self.photometry_iterations_label.grid(row=5, column=0, sticky=tk.W)
        self.photometry_iterations_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.photometry_iterations_entry.grid(row=5, column=1, sticky=tk.E)
        self.set_entry_text(self.photometry_iterations_entry, "1")

        self.sharplo_label = tk.Label(self.settings_frame, text="Lower Bound for Sharpness:")
        self.sharplo_label.grid(row=6, column=0, sticky=tk.W)
        self.sharplo_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.sharplo_entry.grid(row=6, column=1, sticky=tk.E)
        self.set_entry_text(self.sharplo_entry, "0.5")

        self.bkg_filter_size_label = tk.Label(self.settings_frame, text="Background Median Filter, px:")
        self.bkg_filter_size_label.grid(row=7, column=0, sticky=tk.W)
        self.bkg_filter_size_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.bkg_filter_size_entry.grid(row=7, column=1, sticky=tk.E)
        self.set_entry_text(self.bkg_filter_size_entry, "1")

        self.filter_label = tk.Label(self.settings_frame, text="CCD Filter:")
        self.filter_label.grid(row=8, column=0, sticky=tk.W)
        self.filter_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.filter_entry.grid(row=8, column=1, sticky=tk.E)
        self.set_entry_text(self.filter_entry, "")

        self.exposure_label = tk.Label(self.settings_frame, text="Exposure Time:")
        self.exposure_label.grid(row=9, column=0, sticky=tk.W)
        self.exposure_entry = tk.Entry(self.settings_frame, width=settings_entry_width)
        self.exposure_entry.grid(row=9, column=1, sticky=tk.E)
        self.set_entry_text(self.exposure_entry, "0")


        # Here we have full-width settings dropdowns in the right frame
        self.weighting_label = tk.Label(self.right_frame, text="Ensemble Fit Weighting:")
        self.weighting_label.grid(row=3, column=0, sticky=tk.W)
        self.weighting_stringvar = tk.StringVar()
        self.weighting_stringvar.set("None")
        self.weighting_dropdown = tk.OptionMenu(self.right_frame, self.weighting_stringvar, "None", "Raw Flux", "Instrumental Magnitude", "PSF Sigma")
        self.weighting_dropdown.grid(row=4, column=0, sticky=tk.EW)

        self.catalog_label = tk.Label(self.right_frame, text="Comparison Catalog:")
        self.catalog_label.grid(row=5, column=0, sticky=tk.W)
        self.catalog_stringvar = tk.StringVar()
        self.catalog_stringvar.set("APASS DR9")
        self.catalog_dropdown = tk.OptionMenu(self.right_frame, self.catalog_stringvar, "APASS DR9", "URAT1", "USNO-B1.0", "VizieR Catalog")
        self.catalog_dropdown.grid(row=6, column=0, sticky=tk.EW)

        self.vizier_catalog_label = tk.Label(self.right_frame, text="VizieR Catalog Number:")
        self.vizier_catalog_label.grid(row=7, column=0, sticky=tk.W)
        self.vizier_catalog_entry = tk.Entry(self.right_frame)
        self.vizier_catalog_entry.grid(row=8, column=0, sticky=tk.EW)

        # Histogram stretch sliders
        self.stretch_label = tk.Label(self.right_frame, text="Histogram Stretch Low/High:")
        self.stretch_label.grid(row=10, column=0, sticky=tk.W)
        self.stretch_low = tk.Scale(self.right_frame, from_=0, to=100, orient=tk.HORIZONTAL, command=self.update_histogram_low)
        self.stretch_low.grid(row=11, column=0, sticky=tk.EW)
        self.stretch_high = tk.Scale(self.right_frame, from_=0, to=100, orient=tk.HORIZONTAL, command=self.update_histogram_high)
        self.stretch_high.set(5)
        self.stretch_high.grid(row=12, column=0, sticky=tk.EW)


        # Console below
        self.console = tk.Text(self.center, height=10, bg='black', fg='white', width=200)
        self.console.grid(sticky=tk.N+tk.S+tk.E+tk.W, column=0, row=3)
        self.console_scrollbar = tk.Scrollbar(self.center)
        self.console_scrollbar.grid(sticky=tk.N + tk.S + tk.E + tk.W, column=1, row=3)

        self.console.config(yscrollcommand=self.console_scrollbar.set)
        self.console_scrollbar.config(command=self.console.yview)

        self.console_msg("MetroPSF 0.1 Ready.")

        #self.initialize_debug()
        #self.plot_photometry()

        tk.mainloop()


myGUI = MyGUI()




